import type IVue from "vue";

export default function (Vue: typeof IVue) {
  const {
    h,
    ref,
    watch,
    toRef,
    Teleport,
    reactive,
    onMounted,
    onUnmounted,
    defineComponent,
    computed,
  } = Vue;

  return function toVueComponent(
    HFC: typeof HyperFunctionComponent,
    lowercase?: boolean
  ) {
    HFC.propTypes = HFC.propTypes || {};
    const propKeys = Object.keys(HFC.propTypes.attrs || {});
    const eventKeys = Object.keys(HFC.propTypes.events || {});
    const slotKeys = Object.keys(HFC.propTypes.slots || {});

    const propKeyMap: Record<string, string> = {};
    const eventKeyMap: Record<string, string> = {};
    const slotKeyMap: Record<string, string> = {};
    propKeys.forEach((k) => {
      propKeyMap[lowercase ? k.toLowerCase() : k] = k;
    });
    eventKeys.forEach((k) => {
      eventKeyMap[
        lowercase
          ? "on" + k[0].toUpperCase() + k.slice(1).toLowerCase()
          : "on" + k[0].toUpperCase() + k.slice(1)
      ] = k;
    });
    slotKeys.forEach((k) => {
      slotKeyMap[lowercase ? k.toLowerCase() : k] = k;
    });

    const observePropKeys = Object.keys(propKeyMap);
    const observeEventKeys = Object.keys(eventKeyMap);
    const observeSlotKeys = Object.keys(slotKeyMap);

    return defineComponent({
      props: observePropKeys,
      setup(vueProps, { slots: vueSlots, attrs: vueAttrs }) {
        const container = ref<HTMLDivElement>();

        const events: Record<string, any> = {};
        if (observeEventKeys.length) {
          for (let i = 0; i < observeEventKeys.length; i++) {
            const key = observeEventKeys[i];
            const event = vueAttrs[key];
            if (!event) continue;
            events[eventKeyMap[key]] = event;
          }
        }

        const slots: Record<string, any> = {};

        interface RenderSlotItem {
          name: string;
          container: HTMLElement;
          slot: Function;
          args: any;
        }
        const renderedSlots: Record<string, RenderSlotItem> = reactive({});

        if (observeSlotKeys.length) {
          for (let i = 0; i < observeSlotKeys.length; i++) {
            const key = observeSlotKeys[i];
            const slot = vueSlots[key];
            if (!slot) continue;
            slots[slotKeyMap[key]] = function (
              container: HTMLElement | null,
              args: any
            ) {
              if (!container) {
                delete renderedSlots[key];
                return;
              }

              const slotItem = renderedSlots[key];
              if (!slotItem) {
                renderedSlots[key] = { name: key, container, slot, args };
                return;
              }

              if (slotItem.container !== container) {
                slotItem.container = container;
              }

              slotItem.args = args;
            };
          }
        }

        let hfc: HyperFunctionComponent;
        const attrs: Record<string, any> = {};

        if (observePropKeys.length) {
          for (let i = 0; i < observePropKeys.length; i++) {
            const key = observePropKeys[i];
            if (!vueProps[key]) continue;
            attrs[propKeyMap[key]] = vueProps[key];

            watch(
              toRef(vueProps, key),
              (curr, prev) => {
                attrs[propKeyMap[key]] = curr;
                hfc.changed?.("attr", propKeyMap[key], prev, curr);
              },
              { deep: true }
            );
          }
        }

        hfc = new HFC({
          attrs,
          events,
          slots,
        });

        onMounted(() => {
          hfc.connected(container.value as HTMLDivElement);
        });

        onUnmounted(() => {
          hfc.disconnected?.();
        });

        const renderedSlotVnodes = computed(() => {
          return Object.keys(renderedSlots).map((key) => {
            const item = renderedSlots[key];
            return h(Teleport, { to: item.container }, item.slot(item.args));
          });
        });

        return () => [h("div", { ref: container }), renderedSlotVnodes.value];
      },
    });
  };
}
